# Fibonacci Heap in python
# Implemented from https://www.programiz.com/dsa/fibonacci-heap

import math


# Creating fibonacci tree
class FibonacciTree:
    def __init__(self, value):
        self.value = value
        self.child = []
        self.order = 0

    # Adding tree at the end of the tree
    def add_at_end(self, t):
        self.child.append(t)
        self.order = self.order + 1


# Creating Fibonacci heap
class FibonacciHeap:
    def __init__(self):
        self.trees = []
        self.least = None
        self.count = 0

    # Insert a node
    def insert_node(self, value):
        new_tree = FibonacciTree(value)
        self.trees.append(new_tree)
        if self.least is None or value < self.least.value:
            self.least = new_tree
        self.count = self.count + 1

    # Get minimum value
    def get_min(self):
        if self.least is None:
            return None
        return self.least.value

    # Extract the minimum value
    def extract_min(self):
        smallest = self.least
        if smallest is not None:
            for child in smallest.child:
                self.trees.append(child)
            self.trees.remove(smallest)
            if not self.trees:
                self.least = None
            else:
                self.least = self.trees[0]
                self.consolidate()
            self.count = self.count - 1
            return smallest.value

    # Consolidate the tree
    def consolidate(self):
        aux = (floor_log(self.count) + 1) * [None]

        while self.trees:
            x = self.trees[0]
            order = x.order
            self.trees.remove(x)
            while aux[order] is not None:
                y = aux[order]
                if x.value > y.value:
                    x, y = y, x
                x.add_at_end(y)
                aux[order] = None
                order = order + 1
            aux[order] = x

        self.least = None
        for k in aux:
            if k is not None:
                self.trees.append(k)
                if self.least is None or k.value < self.least.value:
                    self.least = k


def floor_log(x):
    return math.frexp(x)[1] - 1


# fibonacci_heap = FibonacciHeap()
#
# fibonacci_heap.insert_node(7)
# fibonacci_heap.insert_node(3)
# fibonacci_heap.insert_node(17)
# fibonacci_heap.insert_node(24)
#
# print('the minimum value of the fibonacci heap: {}'.format(fibonacci_heap.get_min()))
#
# print('the minimum value removed: {}'.format(fibonacci_heap.extract_min()))
#
# import random
# added = {0}
# for i in range(25):
#     n = 0
#     while n in added:
#         n = random.randint(1, 100)
#     fibonacci_heap.insert_node(n)
#     added.add(n)
#
# print('the minimum value of the fibonacci heap: {}'.format(fibonacci_heap.get_min()))
# for i in range(10):
#     print('the minimum value removed: {}'.format(fibonacci_heap.extract_min()))
#     print(len(fibonacci_heap.trees))
# print('the minimum value of the fibonacci heap: {}'.format(fibonacci_heap.get_min()))
#
# print(sorted(added)[:11])
